from lambdalib.lambdalib import Lambda
from lambdalib.stdlib import Inv
from lambdalib.stdlib import Buf
from lambdalib.stdlib import Nand2
from lambdalib.stdlib import Nand3
from lambdalib.stdlib import Nand4
from lambdalib.stdlib import Nor2
from lambdalib.stdlib import Nor3
from lambdalib.stdlib import Nor4


class Ring(Lambda):
    def __init__(self):
        name = 'la_ring'
        super().__init__(name, __file__)

        # deps
        self.add_depfileset(Inv(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Buf(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Nand2(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Nand3(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Nand4(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Nor2(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Nor3(), depfileset='rtl', fileset='rtl')
        self.add_depfileset(Nor4(), depfileset='rtl', fileset='rtl')


if __name__ == "__main__":
    d = Ring()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



from typing import Dict

from siliconcompiler import Design

from .la_asyncfifo.la_asyncfifo import Asyncfifo
from .la_syncfifo.la_syncfifo import Syncfifo
from .la_spram.la_spram import Spram
from .la_spregfile.la_spregfile import Spregfile
from .la_dpram.la_dpram import Dpram
from .la_tdpram.la_tdpram import Tdpram

__all__ = ['Asyncfifo',
           'Syncfifo',
           'Dpram',
           'Spram',
           'Spregfile',
           'Tdpram']


class RAMLib(Design):
    """RAM Library Design containing various RAM and FIFO modules."""
    def __init__(self):
        super().__init__("la_ramlib")

        with self.active_fileset("rtl"):
            self.add_depfileset(Asyncfifo(), depfileset="rtl")
            self.add_depfileset(Syncfifo(), depfileset="rtl")
            self.add_depfileset(Spram(), depfileset="rtl")
            self.add_depfileset(Spregfile(), depfileset="rtl")
            self.add_depfileset(Dpram(), depfileset="rtl")
            self.add_depfileset(Tdpram(), depfileset="rtl")


class RAMTechLib:
    """Abstract base class for RAM technology libraries."""
    def get_ram_width(self) -> int:
        """Returns the width of the RAM cell.

        Returns:
            int: The width of the RAM cell.
        """
        raise NotImplementedError("Subclasses must implement get_ram_width method")

    def get_ram_depth(self) -> int:
        """Returns the depth of the RAM cell.

        Returns:
            int: The depth of the RAM cell.
        """
        raise NotImplementedError("Subclasses must implement get_ram_depth method")

    def get_ram_ports(self) -> Dict[str, str]:
        """Returns the port mapping for the RAM cell.

        Returns:
            Dict[str, str]: A dictionary mapping port names to their expressions.
        """
        raise NotImplementedError("Subclasses must implement get_ram_ports method")

    def get_ram_libcell(self) -> str:
        """Returns the name of the RAM library cell.

        Returns:
            str: The name of the RAM library cell.
        """
        raise NotImplementedError("Subclasses must implement get_ram_libcell method")



from lambdalib.lambdalib import Lambda
from siliconcompiler import Design
from lambdalib.stdlib import Inv
from lambdalib.stdlib import And2
from lambdalib.stdlib import Clkor2
from lambdalib.auxlib import Dsync
from lambdalib.auxlib import Clkicgand


class Clkmux2(Lambda):
    def __init__(self):
        name = 'la_clkmux2'
        super().__init__(name, __file__)

        # dependencies
        deps = [Inv(), And2(), Clkor2(), Dsync(), Clkicgand()]
        with self.active_fileset('rtl'):
            for item in deps:
                self.add_depfileset(item)


class Clkmux2TB(Design):
    def __init__(self):
        name = 'tb_la_clkmux2'
        super().__init__(name)

        # dependencies
        self.set_dataroot(name, __file__)
        with self.active_fileset('rtl'):
            self.add_file(f"testbench/{name}.v")
            self.add_depfileset(Clkmux2())


if __name__ == "__main__":
    d = Clkmux2()
    d.write_fileset(f"{d.name}.f", fileset="rtl")
    d = Clkmux2TB()
    d.write_fileset(f"{d.name}.f", fileset="rtl")



import os.path

from collections import OrderedDict
from jinja2 import Template
from pathlib import Path

from typing import Union, Optional, List, Type, TYPE_CHECKING

from lambdalib.lambdalib import Lambda

if TYPE_CHECKING:
    from lambdalib.ramlib import RAMTechLib


class RAMLib(Lambda):
    def __init__(self, name: str, path: Union[str, Path],
                 impl_file: Optional[Union[str, Path]] = None):
        super().__init__(name, path)

        if impl_file:
            with self.active_dataroot(name), self.active_fileset("rtl.impl"):
                self.add_file(impl_file)
            with self.active_fileset("rtl"):
                self.add_depfileset(self, "rtl.impl")

    def write_lambdalib(self, path: Union[str, Path],
                        memories: List[Type["RAMTechLib"]],
                        min_size: Optional[int] = None) -> None:
        """Writes a fileset file listing the RTL files for this lambda library.

        Args:
            path: The path to the output fileset file.
            memories: A list of memory classes.
            min_size: An optional minimum size for the memories.

        Example:
            >>> lib.write_lambdalib("output.v", lambda_memories.techlibs)
        """
        template_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                     'templates',
                                                     f'{self.name}memory.v'))

        memdata = {}
        for mem in memories:
            memobj = mem()
            memdata[memobj.get_ram_libcell()] = {
                "DW": memobj.get_ram_width(),
                "AW": memobj.get_ram_depth(),
                "port_map": [(port, wire) for port, wire in memobj.get_ram_ports().items()]
            }

        widths_table = []
        depths_table = []
        memory_port_map = {}
        selection_table = {}
        memory_inst_map = {}

        if min_size is None:
            min_size = 0

        for memory, info in sorted(memdata.items(), key=lambda mem: mem[0]):
            widths_table.append(
                (memory, info['DW'])
            )
            depths_table.append(
                (memory, info['AW'])
            )

            memory_port_map[memory] = sorted(info["port_map"])
            memory_inst_map[memory] = memory

            selection_table.setdefault(info['AW'], {})[int(info['DW'])] = memory

        selection_table = OrderedDict(sorted(selection_table.items(), reverse=True))
        for aw, items in selection_table.items():
            selection_table[aw] = OrderedDict(sorted(items.items(), reverse=True))

        widths_table.sort()
        depths_table.sort()

        with open(template_path) as f:
            template = Template(f.read())

        with open(path, 'w') as fout:
            fout.write(template.render(
                type=self.name,
                width_table=widths_table,
                depth_table=depths_table,
                selection_table=selection_table,
                inst_map=memory_inst_map,
                port_mapping=memory_port_map,
                minsize=min_size))



import os
import math
import random
import string
from decimal import Decimal

from pathlib import Path
from typing import List, Tuple, Optional, Mapping, Union

from siliconcompiler import Sim

from cocotb.triggers import Timer
from cocotb.handle import SimHandleBase

from cocotb_tools.runner import get_runner, VerilatorControlFile
from cocotb_tools.check_results import get_results


async def do_reset(
        reset: SimHandleBase,
        time_ns: int,
        active_level: bool = False):
    """Perform a async reset"""
    reset.value = not active_level
    await Timer(1, unit="step")
    reset.value = active_level
    await Timer(time_ns, "ns")
    reset.value = not active_level
    await Timer(1, unit="step")


def run_cocotb(
        project: Sim,
        test_module_name: str,
        output_dir_name: Optional[str] = None,
        simulator_name: str = "icarus",
        build_args: Optional[List] = None,
        timescale: Optional[Tuple[str, str]] = None,
        parameters: Optional[Mapping[str, object]] = None,
        seed: Optional[Union[str, int]] = None,
        waves: bool = True):
    """Launch cocotb given a SC Project"""

    if parameters is None:
        parameters = {}

    if build_args is None:
        build_args = []

    if output_dir_name is None:
        output_dir_name = test_module_name

    pytest_current_test = os.getenv("PYTEST_CURRENT_TEST", None)

    rootpath = Path(__file__).resolve().parent.parent
    top_level_dir = rootpath
    build_dir = rootpath / "build" / output_dir_name
    test_dir = None

    results_xml = None
    if not pytest_current_test:
        results_xml = build_dir / "results.xml"
        test_dir = top_level_dir

    # Get top level module name
    top_lvl_module_name = None
    main_filesets = project.get("option", "fileset")
    if main_filesets and len(main_filesets) != 0:
        main_fileset = main_filesets[0]
        top_lvl_module_name = project.design.get_topmodule(
            fileset=main_fileset
        )

    filesets = project.get_filesets()
    idirs = []
    sc_defines = []
    for lib, fileset in filesets:
        idirs.extend(lib.find_files("fileset", fileset, "idir"))
        sc_defines.extend(lib.get("fileset", fileset, "define"))

    defines = {}
    for define in sc_defines:
        parts = define.split("=", 1)
        if len(parts) == 2:
            defines[parts[0]] = parts[1]
        elif len(parts) == 1:
            defines[parts[0]] = "1"

    sources = []
    for lib, fileset in filesets:
        for value in lib.get_file(fileset=fileset, filetype="systemverilog"):
            sources.append(value)
    for lib, fileset in filesets:
        for value in lib.get_file(fileset=fileset, filetype="verilog"):
            sources.append(value)

    vlt_files = []
    if simulator_name == "verilator":
        for lib, fileset in filesets:
            for value in lib.get_file(fileset=fileset, filetype="verilatorctrlfile"):
                vlt_files.append(VerilatorControlFile(value))

    # Build HDL in chosen simulator
    runner = get_runner(simulator_name)
    runner.build(
        sources=vlt_files + sources,
        includes=idirs,
        hdl_toplevel=top_lvl_module_name,
        build_args=build_args,
        waves=waves,
        timescale=timescale,
        build_dir=build_dir,
        parameters=parameters,
        defines=defines
    )

    # Run test
    _, tests_failed = get_results(runner.test(
        hdl_toplevel=top_lvl_module_name,
        test_module=test_module_name,
        test_dir=test_dir,
        test_args=build_args,
        results_xml=results_xml,
        build_dir=build_dir,
        seed=seed,
        waves=waves
    ))

    return tests_failed


def random_decimal(max: int, min: int, decimal_places=2) -> Decimal:
    prefix = str(random.randint(min, max))
    suffix = ''.join(random.choice(string.digits) for _ in range(decimal_places))
    return Decimal(prefix + "." + suffix)


def random_toggle_generator(on_range=(0, 15), off_range=(0, 15)):
    return bit_toggler_generator(
        gen_on=(random.randint(*on_range) for _ in iter(int, 1)),
        gen_off=(random.randint(*off_range) for _ in iter(int, 1))
    )


def sine_wave_generator(amplitude, w, offset=0):
    while True:
        for idx in (i / float(w) for i in range(int(w))):
            yield amplitude * math.sin(2 * math.pi * idx) + offset


def bit_toggler_generator(gen_on, gen_off):
    for n_on, n_off in zip(gen_on, gen_off):
        yield int(abs(n_on)), int(abs(n_off))


def wave_generator(on_ampl=30, on_freq=200, off_ampl=10, off_freq=100):
    return bit_toggler_generator(sine_wave_generator(on_ampl, on_freq),
                                 sine_wave_generator(off_ampl, off_freq))
